

#pragma once

#include <memory>
#include <cutil_math.h>
#include "GLBuffer.h"


#define NO_ACTIVE_ENTRY 0xFFFFFFFF

struct OctreeEntry
{
  unsigned int vertexOffset;
  unsigned int numVertices;
  unsigned int triangleOffset;
  unsigned int numTriangles;

  #define NO_CHILD_FLAG 0xFFFFFFFF
  #define MAX_DEPTH 32
  __host__ __device__ static  inline bool hasChildren(unsigned entry)
  {
    return entry != 0xFFFFFFFF;
  }
  unsigned int children; //(ccciiiiiiiiiiiiiiiiiiiiiiiiiiiii) 
                          //c is count of children - 1
                          //i index of first child
  #define CHILD_ID_SHIFT 0
  #define CHILD_COUNT_SHIFT 29
  #define CHILD_ID_BITS 0x1FFFFFFF
  #define CHILD_ID_MASK CHILD_ID_BITS
  #define CHILD_COUNT_MASK 0xE0000000
  #define CHILD_COUNT_BITS 0x7
  __host__ __device__ static inline unsigned getNumChildren(unsigned entry)
  {
    return ((entry >> CHILD_COUNT_SHIFT) & CHILD_COUNT_BITS)+1;
  }
  __host__ __device__ static inline unsigned setNumChildren(unsigned num, unsigned entry)
  {
    return (((num-1) & CHILD_COUNT_BITS) << CHILD_COUNT_SHIFT) | (entry & CHILD_ID_BITS);
  }
  __host__ __device__ static inline unsigned getFirstChild(unsigned entry)
  {
    return (entry >> CHILD_ID_SHIFT) & CHILD_ID_BITS;
  }
  __host__ __device__ static inline unsigned setFirstChild(unsigned child, unsigned entry)
  {
    return (entry & CHILD_COUNT_MASK) | ((child & CHILD_ID_BITS) << CHILD_ID_SHIFT);
  }
  OctreeEntry()
  {
    vertexOffset = numVertices = triangleOffset = numTriangles = 0;
    children = NO_CHILD_FLAG;
  }

};


struct CUDAStorages
{
  GL::Buffer vertex_buffer;
  GL::Buffer normal_buffer;
  GL::Buffer index_buffer;


  cuda_ptr<OctreeEntry> d_octree_data;
  cuda_ptr<unsigned int> d_octree_vertex_coherence;
  cuda_ptr<unsigned int> d_octree_triangle_coherence;
  cuda_ptr<unsigned int> d_vertex_lookup;

  size_t active_set;
  unsigned int active_nodes_count;
  cuda_ptr<unsigned int> d_active_set[2];


  size_t vertexcount;
  size_t trianglecount;
  size_t voxelcount;
  size_t leafs;
  size_t maxdepth;

  std::unique_ptr<unsigned int[]> h_levelOffset;
  cuda_ptr<uint3> d_voxel_ids;

  size_t reductionClusterSize;

  size_t triangle_coherence_size;
  size_t vertex_coherence_size;

  float boundaries[6];

  CUDAStorages();
  ~CUDAStorages();
};

