

  __constant__ OctreeEntry* octree;
  __constant__ unsigned int *vertexIdOut, *triOutput;
  __constant__ const unsigned int *vertexCoherence, *triangleCoherence;
  __constant__ unsigned int vertices, maxdepth;
  __constant__ const uint3* voxelIds;


  __device__ uint mark(uint node)
  {
    return 0x80000000 | node;
  }
  __device__ uint unmark(uint node)
  {
    return (~0x80000000) & node;
  }
  __device__ bool isMarked(uint node)
  {
    return ((node & 0x80000000) == 0x80000000)?true:false;
  }

  template<bool DEPTHERROR>
  __global__ void checkLevel(uint* activeSet, uint* numberofchildren, uint* numberOfLeafs, uint depth, uint num, float threshold)
  {
    uint id = blockDim.x * blockIdx.x + threadIdx.x;
    if(id >= num) return;
    uint mynodeid = activeSet[id];
    uint childnodes = 0, leafnodes = 0;

    const OctreeEntry* node = octree + mynodeid;
    uint numChildren = OctreeEntry::hasChildren(node->children)?OctreeEntry::getNumChildren(node->children):0;
    uint firstChildId = OctreeEntry::getFirstChild(node->children);
    for(uint i = 0; i < numChildren; ++i)
    {
      uint thisChildId = firstChildId + i;
      const OctreeEntry *childEntry = octree + thisChildId;
      //get voxel boundaries
      float3 vMin, vMax;
      getVoxelBounds(voxelIds[thisChildId], depth+1, vMin, vMax);
      //check if criterion is fullfilled
      bool expand = carryOnTraversal<DEPTHERROR>(depth+1, vMin, vMax, threshold);
      if(expand && (depth+1) < maxdepth)
        ++childnodes;
      else //leaf
        ++leafnodes;
    }
    if(numChildren == 0) 
       ++leafnodes;

    numberofchildren[id] = childnodes;
    numberOfLeafs[id] = leafnodes;
  }

 template<bool DEPTHERROR>
  __global__ void writeLevel(const uint* activeSet, const uint* numberofchildren, const  uint* numberOfLeafs, uint* children, uint* childTriangles, uint* leafs, uint depth, uint num, float threshold)
  {
    uint id = blockDim.x * blockIdx.x + threadIdx.x;
    if(id >= num) return;
    uint mynodeid = activeSet[id];
    uint childnodes_offset = numberofchildren[id];
    uint leafnodes_offset = numberOfLeafs[id];

    const OctreeEntry* node = octree + mynodeid;
    uint numChildren = OctreeEntry::hasChildren(node->children)?OctreeEntry::getNumChildren(node->children):0;
    uint firstChildId = OctreeEntry::getFirstChild(node->children);
    for(uint i = 0; i < numChildren; ++i)
    {
      uint thisChildId = firstChildId + i;
      const OctreeEntry *childEntry = octree + thisChildId;
      //get voxel boundaries
      float3 vMin, vMax;
      getVoxelBounds(voxelIds[thisChildId], depth+1, vMin, vMax);
      //check if criterion is fullfilled
      bool expand = carryOnTraversal<DEPTHERROR>(depth+1, vMin, vMax, threshold);
      if(expand && (depth+1) < maxdepth)
      {
        children[childnodes_offset] = thisChildId;
        childTriangles[childnodes_offset++] = childEntry->numTriangles;
      }
      else //leaf
        leafs[leafnodes_offset++] = thisChildId;
    }
    if(numChildren == 0) 
       leafs[leafnodes_offset++] = (depth + 1 >= maxdepth)?mark(mynodeid):mynodeid;
  }

  __global__ void handleLeafs(const uint* leafs,  uint num)
  {
    uint id = blockDim.x * blockIdx.x + threadIdx.x;
    if(id >= num) return;
    uint nodeId = leafs[id];
    bool marked = isMarked(nodeId);
    nodeId = unmark(nodeId);
    const OctreeEntry *node = octree + nodeId;

    uint vertexTargetStart = node->vertexOffset;
    uint vertexTargetEnd = vertexTargetStart + node->numVertices;
    for(uint targetVertex = vertexTargetStart; targetVertex < vertexTargetEnd; ++targetVertex)
    {
      uint coherenceId = vertexCoherence[targetVertex];
      vertexIdOut[coherenceId] = marked?coherenceId:(vertices + nodeId);
    }
  }

  __global__ void handleNodes(const uint* nodes, const uint* triangleOffset, uint num)
  {
    uint id = blockDim.x * blockIdx.x + threadIdx.x;
    if(id >= num) return;
    uint nodeId = nodes[id];
    uint triOffset = triangleOffset[id];
    const OctreeEntry *node = octree + nodeId;

    uint numTris = node->numTriangles;
    if(numTris == 0)
      return;

    uint startWrite = triOffset;
    uint * indexoutput = triOutput + 3*startWrite;


    //run through all triangles
    for(uint i = triOffset; i < triOffset + numTris; ++i)
    {
      //all 3 vertices
      for(unsigned int v = 0; v < 3; ++v)
        *indexoutput++ = vertexIdOut[triangleCoherence[3*i + v]];
    }
  }