#include <cuda.h>
#include <cutil.h>
#include <cutil_math.h>

#include "scene.h"
#include "structures.h"

#include <cuda.h>
#include "renderingdata.cuh"
#include "corefunctions.cuh"

#include "timing.h"



__global__ void startTrace(rgbcolor* imagedata, uint* randseeds);


double traceCuda(int device, const Scene& scene)
{
  cudaSetDevice(device);
  setupData(scene);
  initImage();
  initRand();

  double t0 = Timing::gettime();
  startTrace<<<scene.imgDims(), tsamples>>>(d_imagedata, d_randseeds);
  cudaDeviceSynchronize();
  double t1 = Timing::gettime();
  return t1 - t0;
}




__global__ void startTrace(rgbcolor* imagedata, uint* randseeds)
{
  uint2 pixel = make_uint2( blockIdx.x, blockIdx.y);
	unsigned int randseed =  getSeed(randseeds, pixel, threadIdx.x);

  vector3 contrib = make_float3(1.0f);
  vector3 orig = make_float3(0.0f);
  rgbcolor color = make_float3(0.0f);

  vector3 dir;
  dir.x = ( (float)pixel.x / (float)c_width ) - 0.5f + device_random(randseed)/(float)c_width;
	dir.y = ( ( (float)pixel.y / (float)c_height ) - 0.5f ) * ( (float)c_height/(float)c_width) + device_random(randseed)/(float)c_height;
	dir.z = 1;
	dir = vec_normalize( dir );


  uint recursion = 0;
  while( ( recursion < c_maxrecursions ) && ( ( contrib.x > MIN_CONTRIB ) || ( contrib.y  > MIN_CONTRIB ) || ( contrib.z > MIN_CONTRIB ) ) )
  {
    float3 d, n;
    float t;
    const object* obj = intersectObjects(c_objects, c_numObjects, d, n, t, orig, dir);

    if(obj) 
      hitObject(*obj, t, n, orig, dir, contrib, color, randseed);
    else
		{
      color += c_background * contrib;
		  contrib.x = contrib.y = contrib.z = 0;
		}
    recursion++;
	}

  rgbcolor* outcolor = imagedata + (pixel.y*c_width + pixel);
  atomicAdd(&outcolor->x, color.x);
  atomicAdd(&outcolor->y, color.y);
  atomicAdd(&outcolor->z, color.z);
}